import os
import pytest

# import shutil

# from pathlib import Path

from cellfinder.tools import prep
from imlib.general.exceptions import CommandLineInputError


data_dir = os.path.join("tests", "data")


def test_check_return_ch_ids():
    signal_ch = [0, 1, 3]
    bg_ch = 6
    signal_list = ["file1.txt", "file_2.txt", "file_3.txt"]
    # None given
    assert ([0, 1, 2], 3) == prep.check_and_return_ch_ids(
        None, None, signal_list
    )
    # Only signal given
    assert (signal_ch, 4) == prep.check_and_return_ch_ids(
        signal_ch, None, signal_list
    )

    # Only background given
    assert ([7, 8, 9], bg_ch) == prep.check_and_return_ch_ids(
        None, bg_ch, signal_list
    )

    # Both given (no overlap)
    assert (signal_ch, bg_ch) == prep.check_and_return_ch_ids(
        signal_ch, bg_ch, signal_list
    )

    # Both given (overlap)
    with pytest.raises(CommandLineInputError):
        assert prep.check_and_return_ch_ids(signal_ch, 3, signal_list)


class Args:
    def __init__(
        self,
        model_dir=None,
        empty=None,
        no_detection=False,
        no_classification=False,
        no_register=False,
        no_analyse=False,
        no_standard_space=False,
        output_dir=None,
        registration_output_folder=None,
        cells_file_path=None,
        cubes_output_dir=None,
        classification_out_file=None,
        cells_in_standard_space=None,
    ):
        self.cell_count_model_dir = model_dir
        self.empty = empty

        self.no_detection = no_detection
        self.no_classification = no_classification
        self.no_register = no_register
        self.no_summarise = no_analyse
        self.no_standard_space = no_standard_space

        self.output_dir = output_dir

        self.paths = Paths(
            output_dir=registration_output_folder,
            cells_file_path=cells_file_path,
            cubes_output_dir=cubes_output_dir,
            classification_out_file=classification_out_file,
            cells_in_standard_space=cells_in_standard_space,
        )


class Paths:
    def __init__(
        self,
        output_dir=None,
        cells_file_path=None,
        cubes_output_dir=None,
        classification_out_file=None,
        cells_in_standard_space=None,
    ):
        self.registration_output_folder = output_dir
        self.cells_file_path = cells_file_path
        self.tmp__cubes_output_dir = cubes_output_dir
        self.classification_out_file = classification_out_file
        self.cells_in_standard_space = cells_in_standard_space


# def get_dict_of_what_to_run(what_to_run):
#     what_to_run_dict = {
#         "detect": what_to_run.detect,
#         "classify": what_to_run.classify,
#         "register": what_to_run.register,
#         "figures": what_to_run.figures,
#     }
#     return what_to_run_dict
#
#
# def test_calc_what_to_run(tmpdir):
#     tmpdir = str(tmpdir)
#
#     # registered_atlas_path = os.path.join(tmpdir, "registered_atlas.tiff")
#     cells_file_path = os.path.join(tmpdir, "cells.xml")
#     classification_out_file_path = os.path.join(
#         tmpdir, "cell_classification.xml"
#     )
#     summary_cell_counts_path = os.path.join(
#     tmpdir, "summary_cell_counts.csv")
#     figures_dir = os.path.join(tmpdir, "figures")
#
#     args = Args(
#         output_dir=tmpdir,
#         registration_output_folder=tmpdir,
#         cells_file_path=cells_file_path,
#         classification_out_file=classification_out_file_path,
#         figures_dir=figures_dir,
#     )
#
#     # default
#     what_to_run = prep.CalcWhatToRun(args)
#     default_validation = {
#         "detect": True,
#         "classify": True,
#         "register": False,
#         "figures": False,
#     }
#     default_test = get_dict_of_what_to_run(what_to_run)
#     assert default_validation == default_test
#
#     # options
#     args.no_classification = True
#     what_to_run = prep.CalcWhatToRun(args)
#     validation = {
#         "detect": True,
#         "classify": False,
#         "register": False,
#         "figures": False,
#     }
#     test = get_dict_of_what_to_run(what_to_run)
#     assert validation == test
#
#
#     # # existance
#     # Path(registered_atlas_path).touch()
#     # what_to_run = prep.CalcWhatToRun(args)
#     # validation = {
#     #     "detect": True,
#     #     "classify": True,
#     #     "register": False,
#     #     "figures": True,
#     # }
#     # test = get_dict_of_what_to_run(what_to_run)
#     # assert validation == test
#
#     Path(cells_file_path).touch()
#     what_to_run.update(args)
#     validation = {
#         "detect": False,
#         "classify": True,
#         "register": False,
#         "figures": True,
#     }
#     test = get_dict_of_what_to_run(what_to_run)
#     assert validation == test
#
#     os.remove(cells_file_path)
#     what_to_run.update(args)
#     validation = {
#         "detect": True,
#         "classify": True,
#         "register": False,
#         "figures": True,
#     }
#     test = get_dict_of_what_to_run(what_to_run)
#     assert validation == test
#
#     Path(classification_out_file_path).touch()
#     what_to_run.update(args)
#     validation = {
#         "detect": False,
#         "classify": False,
#         "register": False,
#         "figures": True,
#     }
#     test = get_dict_of_what_to_run(what_to_run)
#     assert validation == test
#
#     Path(summary_cell_counts_path).touch()
#     what_to_run.update(args)
#     validation = {
#         "detect": False,
#         "classify": False,
#         "register": False,
#         "summarise": False,
#         "figures": True,
#     }
#     test = get_dict_of_what_to_run(what_to_run)
#     assert validation == test
#
#     make_and_fill_directory(figures_dir)
#     args.figures = True
#     what_to_run.update(args)
#     validation = {
#         "detect": False,
#         "classify": False,
#         "register": False,
#         "summarise": False,
#         "figures": False,
#     }
#     test = get_dict_of_what_to_run(what_to_run)
#     assert validation == test
#
#
#     # os.remove(registered_atlas_path)
#     os.remove(summary_cell_counts_path)
#     shutil.rmtree(figures_dir)
#     what_to_run.update(args)
#     validation = {
#         "detect": False,
#         "classify": False,
#         "register": True,
#         "summarise": True,
#         "figures": True,
#     }
#     test = get_dict_of_what_to_run(what_to_run)
#     assert validation == test


def make_and_fill_directory(directory):
    os.mkdir(directory)
    for file_size in range(100, 200, 20):
        with open(os.path.join(directory, str(file_size)), "wb") as fout:
            fout.write(os.urandom(file_size))



import logging

import pandas as pd


def export_points(
    points,
    resolution,
    output_filename,
):
    logging.info("Exporting to brainrender")
    points = pd.DataFrame(points * resolution)
    points.columns = ["x", "y", "z"]
    points.to_hdf(output_filename, key="df", mode="w")



"""
main
===============

Runs each part of the cellfinder pipeline in turn.

N.B imports are within functions to prevent tensorflow being imported before
it's warnings are silenced
"""

import os
import logging
import tifffile
from datetime import datetime
import bg_space as bgs
from imlib.general.logging import suppress_specific_logs

tf_suppress_log_messages = [
    "multiprocessing can interact badly with TensorFlow"
]


def get_downsampled_space(atlas, downsampled_image_path):
    target_shape = tifffile.imread(downsampled_image_path).shape
    downsampled_space = bgs.AnatomicalSpace(
        atlas.metadata["orientation"],
        shape=target_shape,
        resolution=atlas.resolution,
    )
    return downsampled_space


def main():
    suppress_tf_logging(tf_suppress_log_messages)
    from brainreg.main import main as register

    from cellfinder.tools import prep

    start_time = datetime.now()
    args, arg_groups, what_to_run, atlas = prep.prep_cellfinder_general()

    if what_to_run.register:
        # TODO: add register_part_brain option
        logging.info("Registering to atlas")
        args, additional_images_downsample = prep.prep_registration(args)
        register(
            args.atlas,
            args.orientation,
            args.target_brain_path,
            args.brainreg_paths,
            args.voxel_sizes,
            arg_groups["NiftyReg registration backend options"],
            sort_input_file=args.sort_input_file,
            n_free_cpus=args.n_free_cpus,
            additional_images_downsample=additional_images_downsample,
            backend=args.backend,
            debug=args.debug,
        )

    else:
        logging.info("Skipping registration")

    if len(args.signal_planes_paths) > 1:
        base_directory = args.output_dir

        for idx, signal_paths in enumerate(args.signal_planes_paths):
            channel = args.signal_ch_ids[idx]
            logging.info("Processing channel: " + str(channel))
            channel_directory = os.path.join(
                base_directory, "channel_" + str(channel)
            )
            if not os.path.exists(channel_directory):
                os.makedirs(channel_directory)

            # prep signal channel specific args
            args.signal_planes_paths[0] = signal_paths
            # TODO: don't overwrite args.output_dir - use Paths instead
            args.output_dir = channel_directory
            args.signal_channel = channel
            # Run for each channel
            run_all(args, what_to_run, atlas)

    else:
        args.signal_channel = args.signal_ch_ids[0]
        run_all(args, what_to_run, atlas)
    logging.info(
        "Finished. Total time taken: {}".format(datetime.now() - start_time)
    )


def run_all(args, what_to_run, atlas):
    from cellfinder.detect import detect
    from cellfinder.classify import classify
    from cellfinder.analyse import analyse
    from cellfinder.figures import figures

    from cellfinder.tools import prep

    args, what_to_run = prep.prep_channel_specific_general(args, what_to_run)

    if what_to_run.detect:
        logging.info("Detecting cell candidates")
        args = prep.prep_candidate_detection(args)
        detect.main(args)
    else:
        logging.info("Skipping cell detection")

    if what_to_run.classify:
        args = prep.prep_classification(args, what_to_run)
        if what_to_run.classify:
            logging.info("Running cell classification")
            what_to_run.cells_exist = classify.main(args)

        else:
            logging.info("No cells were detected, skipping classification.")

    else:
        logging.info("Skipping cell classification")

    what_to_run.update_if_cells_required()

    if what_to_run.analyse or what_to_run.figures:
        downsampled_space = get_downsampled_space(
            atlas, args.brainreg_paths.boundaries_file_path
        )

    if what_to_run.analyse:
        logging.info("Analysing cell positions")
        analyse.run(args, atlas, downsampled_space)
    else:
        logging.info("Skipping cell position analysis")

    if what_to_run.figures:
        logging.info("Generating figures")
        figures.run(args, atlas, downsampled_space.shape)
    else:
        logging.info("Skipping figure generation")


def suppress_tf_logging(tf_suppress_log_messages):
    """
    Prevents many lines of logs such as:
    "2019-10-24 16:54:41.363978: I tensorflow/stream_executor/platform/default
    /dso_loader.cc:44] Successfully opened dynamic library libcuda.so.1"
    """
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

    for message in tf_suppress_log_messages:
        suppress_specific_logs("tensorflow", message)


if __name__ == "__main__":
    main()



import os
import sys
import platform
import pytest
import numpy as np
import pandas as pd

from imio.load import load_any

from cellfinder.main import main as cellfinder_run

data_dir = os.path.join(
    os.getcwd(),
    "tests",
    "data",
    "brain",
)

test_niftyreg_output = os.path.join(
    os.getcwd(), "tests", "data", "registration_output", platform.system()
)

x_pix = "40"
y_pix = "40"
z_pix = "50"

relative_tolerance = 0.01
absolute_tolerance = 10
check_less_precise_pd = 1


@pytest.mark.xfail(reason="Issues across machines")
@pytest.mark.slow
def test_registration_niftyreg(tmpdir):
    output_directory = os.path.join(str(tmpdir), "output")
    cellfinder_args = [
        "cellfinder",
        "-s",
        data_dir,
        "-b",
        data_dir,
        "-o",
        output_directory,
        "-v",
        z_pix,
        y_pix,
        x_pix,
        "--orientation",
        "psl",
        "--n-free-cpus",
        "0",
        "--atlas",
        "allen_mouse_100um",
        "--no-detection",
        "--no-classification",
        "--no-analyse",
        "--no-figures",
    ]

    sys.argv = cellfinder_args
    cellfinder_run()

    # none of this testing is ideal, as results seem to vary between systems

    if platform.system() == "Linux":
        image_list = [
            "boundaries.tiff",
            "deformation_field_0.tiff",
            "deformation_field_1.tiff",
            "deformation_field_2.tiff",
            "downsampled.tiff",
            "downsampled_channel_0.tiff",
            "downsampled_standard.tiff",
            "downsampled_standard_channel_0.tiff",
            "registered_atlas.tiff",
            "registered_hemispheres.tiff",
        ]
    else:
        image_list = [
            "boundaries.tiff",
            "deformation_field_0.tiff",
            "deformation_field_1.tiff",
            "deformation_field_2.tiff",
            "downsampled.tiff",
            "downsampled_channel_0.tiff",
            # "downsampled_standard.tiff",
            # "downsampled_standard_channel_0.tiff",
            # "registered_atlas.tiff",
            # "registered_hemispheres.tiff",
        ]
    output_directory = os.path.join(output_directory, "registration")
    for image in image_list:
        are_images_equal(image, output_directory, test_niftyreg_output)

    if platform.system() == "Linux":
        pd.testing.assert_frame_equal(
            pd.read_csv(os.path.join(output_directory, "volumes.csv")),
            pd.read_csv(os.path.join(test_niftyreg_output, "volumes.csv")),
        )


def are_images_equal(image_name, output_directory, test_output_directory):
    image = load_any(
        os.path.join(output_directory, image_name),
    )
    test_image = load_any(
        os.path.join(test_output_directory, image_name),
    )
    np.testing.assert_allclose(
        image, test_image, rtol=relative_tolerance, atol=absolute_tolerance
    )
